@@ -4,8 +4,7 @@ use subscript_compiler;
 
 fn main() {
     let source = include_str!("./other/bad-path.txt");
-
-    let parsed = subscript_compiler::frontend::run_compiler_frontend(source);
+    let parsed = subscript_compiler::frontend::pass::pp_normalize::run_compiler_frontend(source);
     for node in parsed {
         println!("{:#?}", node);
     }
